#include "Game.h"

bool Game::debug = false;
CL_Font *Game::font;

// Constructor
Game::Game() {

	keys[0].set(CL_KEY_D); 
	keys[1].set(CL_KEY_F); 
	keys[2].set(CL_KEY_R);
	keys[3].set(CL_KEY_P);

	ship = new Ship;
	asteroids = new Asteroids;
	powerups = new Special;
	ufos = new UFO;

	font = new CL_Font("Fonts/font", resourceManager);
	gameLevel = 1; score = 0; accuracyRounds = 0;

	ship->setGameStatePointer(&currentState);

	ufos->shipToFireAt = ship;
	powerups->gameShip = ship;

	message.setFont(font);
	message.setLocation(SCREENWIDTH/2, SCREENHEIGHT/2);

	debugShowFPS = false;
	paused = false;
	powerupsOn = true;
	doublingPoints = false;

	pointsDoubled = 0;
	levelStartingDelay = 3;
	levelCompleteDelay = 5;
	respawnDelay = 4;
	ufoDelay = 10;

	currentState = STATE_LEVEL_STARTING;

	ship->startShip();
	newLevel();
        cout << "Created game" << endl;
}

// Destructor
Game::~Game() {

	cout << "Deleting worldobjects\n";

	delete ship;
	delete asteroids;
	delete powerups;
	delete ufos;
	delete font;
}

void Game::run() {

	initGL();
	resizeGLScene();

        cout << "done gl" << endl;
	// Start timescaling variable
	lastTimerUpdate = CL_System::get_time();

// =============================================================================

	// =========================================================================
	// Game Loop
	// =========================================================================
	while (active && CL_Keyboard::get_keycode(CL_KEY_ESCAPE) == false) {

		// Update timescale object
		unsigned int currentTimerUpdate = CL_System::get_time();

		float fractionOfSecond;
		
		if (lastTimerUpdate - currentTimerUpdate == 0) {
			fractionOfSecond = 0;
		} else {
			fractionOfSecond = (float(currentTimerUpdate) - float(lastTimerUpdate)) / 1000.0f;
		}

		lastTimerUpdate = currentTimerUpdate;

		keepAlive();

                //CL_Display::draw_line(0,0,SCREENHEIGHT,SCREENWIDTH, CL_Color::blue);
		paint();
		// Check key states
		handleKey();

		message.update(fractionOfSecond);
		if (!message.isExpired()) message.paint();

		if (paused) continue;

		switch (currentState) {

			case STATE_LEVEL_STARTING: 

				ship->timeElapsing = false;
				ship->update(fractionOfSecond);
				levelStartingDelay -= fractionOfSecond;

				if (levelStartingDelay < 0) {
					currentState = STATE_LEVEL_PLAYING;
					levelStartingDelay = 3;
				}

				if (!ship->isExpired()) detectCollisions();

				break;

			case STATE_LEVEL_PLAYING:

				ship->timeElapsing = false;
				ship->update(fractionOfSecond);
				asteroids->update(fractionOfSecond);

				if (powerupsOn) { 
					powerups->update(fractionOfSecond);
					powerups->paint();

					ufos->update(fractionOfSecond);

					if (ufoDelay < 0) {
						ufos->setLocationAndPath();
						ufoDelay = 10 + rand() % 10;
					} else if (ufos->isExpired()) {
						ufoDelay -= fractionOfSecond;
					}
				}

				if (!ship->isExpired()) detectCollisions();

				break;

			case STATE_LEVEL_COMPLETE: 

				ship->timeElapsing = false;
				ship->update(fractionOfSecond);
				levelCompleteDelay -= fractionOfSecond;

				if (powerupsOn) 
					ufos->update(fractionOfSecond);

				if (levelCompleteDelay < 0) {
					currentState = STATE_LEVEL_STARTING;
					levelCompleteDelay = 3;
					newLevel();
				}

				break;

			case STATE_RESPAWNING: case STATE_GAME_OVER:
				
				asteroids->update(fractionOfSecond);
				ship->exhaust.update(fractionOfSecond);
				ship->explosionParticles.update(fractionOfSecond);

				if (powerupsOn)
					ufos->update(fractionOfSecond);

				if (currentState == STATE_GAME_OVER) break;

				respawnDelay -= fractionOfSecond;

				if (respawnDelay < 0) {
				
					currentState = STATE_LEVEL_PLAYING;
					ship->startShip();
					respawnDelay = 3;
				}

				break;

			case STATE_PAUSED: 
				// Do nothing
				break;
		}
	}
}

void Game::paint() {

        WorldObject::glState->set_active();

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);	// Clear Screen And Depth Buffer
	
	std::string statusString("Asteroid Arena\n");
        std::ostringstream oss;
	statusString += "Level: ";
        oss<<gameLevel;
	statusString += oss.str();
	statusString += "\n";

	statusString += "Score: ";
        oss.str("");
        oss<<score;
	statusString += oss.str();
	statusString += "\n";

	if (debugShowFPS) {
		statusString += "FPS: "; statusString += fps.update();
	} else 
		fps.update();

	font->draw(6, 6, statusString);

	asteroids->paint();

	paintAccuracyRounds();
	paintLivesRemaining();

	ufos->paint();
	ship->paint();
}

void Game::initGL() {

        WorldObject::glState->set_active();
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glClearColor(0, 0, 0, 0); //set the background colour
	glViewport(0, 0, SCREENWIDTH, SCREENHEIGHT); //let GL know the size of the window
	glShadeModel(GL_FLAT); //set the shading mode
	glDisable(GL_DEPTH_TEST); //everything's flat, so no need for depth testing
	glColor3f(1, 1, 1);

}

void Game::resizeGLScene() {
        WorldObject::glState->set_active();
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	// This is the only remarkable openGL line - it sets the display mode to 'orthogonal' so we can treat 
	// everything as just a flat surface.
	glOrtho(0,SCREENWIDTH, SCREENHEIGHT,0,-1,1);
}

void Game::detectCollisions()
{

	SimpleList<Asteroid> *asteroidsList = asteroids->getAsteroidsList();
	SimpleList<Bullet> *bulletsList = ship->getBulletsList();

	// ==========================
	// Detect collisions
	// ==========================

	// Ship collision with asteroids
	asteroidsList->start();

	while (asteroidsList->next()) {
		
		if (!asteroidsList->get()->isExpired() && asteroidsList->get()->collides(ship->bound)) {

			// If the ship hit the asteroid, figure the amount of deflection the ship will have
			Asteroid *asteroidHit = asteroidsList->get();

			int deflection = 0;

			switch (asteroidHit->asteroidSize) {
				case LARGE: deflection = 300; break;
				case MED: deflection = 200; break;
				case SMALL: deflection = 100; break;
			}

			if (currentState != STATE_LEVEL_STARTING) {
				// If the game has started, apply the hits to the involved atseroid and ship
				asteroidHit->isHit();
				ship->isHit();
				asteroids->asteroidHit(asteroidHit, ship, gameLevel);
			} else {
				// If the game hasn't officially started, no damage occurs
				deflection /= 4;
			}

			ship->bounceOffObject(asteroidHit, deflection);
		}
	}
	
	// Bullet collision with asteroids
	asteroidsList->start();

	while (asteroidsList->next()) {

		bulletsList->start();

		while (bulletsList->next() && !asteroidsList->get()->isExpired()) {

			if (!bulletsList->get()->isExpired() && asteroidsList->get()->collides(bulletsList->get()->bound)) {
	
				bulletsList->get()->isHit();
				asteroidsList->get()->isHit();
				asteroids->asteroidHit(asteroidsList->get(), bulletsList->get(), gameLevel);
				ship->bulletHitSomething();

				int multiplier = 1;

				if (doublingPoints) {
					multiplier++;
					pointsDoubled++;

					if (pointsDoubled > 5) {
						doublingPoints = false;
						pointsDoubled = 0;
					}
				}

				switch (asteroidsList->get()->asteroidSize) {

					case LARGE: score += (100*multiplier); break;
					case MED: score += (200*multiplier); break;
					case SMALL: score += (400*multiplier); break;
				}
			}
		}
	}

	// Bullet collision with ufos
	bulletsList->start();

	while (bulletsList->next() && !ufos->isExpired()) {
		if (!bulletsList->get()->isExpired() && ufos->collides(bulletsList->get()->bound)) {
			ufos->isHit();
			ship->bulletHitSomething();
			doublingPoints ? score += 800 : score += 400;
		}
	}

	// Ufo bullet collision with ship
	ufos->bulletsFired.start();

	while (ufos->bulletsFired.next()) {
		if (ufos->bulletsFired.get()->collides(ship->bound)) {
			ufos->bulletsFired.get()->isHit();
			ship->isHit();
			ship->bounceOffObject(ufos->bulletsFired.get(), 200);
		}
	}

	// Ship collision with ufos
	if (ufos->collides(ship->bound)) {
		ufos->isHit();
		ship->isHit();
		ship->bounceOffObject(ufos, 250);
	}

	// Powerup collision + return code for new powerup if hit
	int powerup = powerups->checkHit();

	if (powerup != -1) {

		// Perform powerups
		switch (powerup) {

			case POWERUP_INVULNERABLE:
				message.setMessage(std::string("Invulnerability"));
				message.setTimeRemaining(1.5f);
				ship->shieldPowerup.active = true;
				ship->shieldPowerup.lengthElapsed = 0;
				break;
			case POWERUP_SHIELD: 
				message.setMessage("Shields Boosted");
				message.setTimeRemaining(1.5f);
				if (ship->shieldPower < .8f) ship->shieldPower += .2f;
				break;
			case POWERUP_GUN: 
				message.setMessage("Cannons Added");
				message.setTimeRemaining(1.5f);
				ship->gunPowerup.active = true;
				ship->gunPowerup.lengthElapsed = 0;
				break;
			case POWERUP_LIFE: 
				message.setMessage("Extra Life");
				message.setTimeRemaining(1.5f);
				ship->lives++;
				break;
			case POWERUP_NONE: 
				message.setMessage("Empty!");
				message.setTimeRemaining(1.5f);
				break;
			case POWERUP_MINE: 
				ship->bounceOffObject(powerups, 350);
				powerups->explosionParticles.burst(80);
				break;
			case POWERUP_POINTDOUBLE:
				message.setMessage("Point 2x Multiplier");
				message.setTimeRemaining(1.5f);
				doublingPoints = true;
				break;
		}
	}

	// Ship died from a hit
	if (ship->isExpired()) {

		// Has more lives
		if (ship->lives > 0) {
			message.setMessage(std::string("Respawning..."));
			message.setTimeRemaining(respawnDelay);
			currentState = STATE_RESPAWNING;
		} 
		// Game over
		else {
			message.setMessage(std::string("Game Over"));
			message.doesNotExpire = true;
			currentState = STATE_GAME_OVER;
		}

		ufos->reset();
		powerups->reset();
	}

	// Level done
	if (asteroidsList->getSize() == 0) {

		// Level-end chores
		gameLevel++;
		ship->clearBulletList();
		ufos->reset();
		powerups->reset();

		currentState = STATE_LEVEL_COMPLETE;

		std::string startMessage("Level Complete");
		message.setMessage(startMessage);
		message.setTimeRemaining(levelCompleteDelay);

		// Check for accuracy round
		float accuracy = float(ship->hits)/float(ship->shots);

		if (accuracy > .75f) {
			accuracyRounds++;
			startMessage += "\nAccuracy Round!";
			message.setMessage(startMessage);
		}
	}
}

void Game::handleKey() {

	//cout << "Handling Key\n";

	// 'D'
	if (keys[0].isPressed())
		debug ? debug = false: debug = true;

	// 'F'
	if (keys[1].isPressed())
		debugShowFPS ? debugShowFPS = false: debugShowFPS = true;

	// 'R'
	if (keys[2].isPressed())
		restartGame();

	// 'P'
	if (keys[3].isPressed()) {
		if (paused) {
			paused = false;
			message.doesNotExpire = false;
		} else {
			paused = true;
			message.setMessage(std::string("Paused"));
			message.doesNotExpire = true;
		}
	}
}

void Game::newLevel() {

	// Level start (breathing time)
        std::ostringstream oss;
        oss << gameLevel;
	std::string startMessage("Level: "); startMessage += oss.str();
	message.setMessage(startMessage);
	message.setTimeRemaining(levelStartingDelay);
	message.doesNotExpire = false;

	// Level score bonus
	score += (gameLevel-1) * 1000;
	ship->shieldPower += .2f;
	if (ship->shieldPower > 1) ship->shieldPower = 1;

	// Add the requisite number of asteroids
	for (int i=0; i < gameLevel+2; i++)
		asteroids->addAsteroid(ship);

	ship->hits = 0;
	ship->shots = 0;
	ship->timeElapsed = 0;
	powerups->reset();

	ufoDelay = 5 + rand() % 10;
}

void Game::restartGame() {

	// Reset tracking variables
	gameLevel = 1;
	score = 0;
	accuracyRounds = 3;

	// Reset delays
	levelStartingDelay = 3;
	levelCompleteDelay = 5;
	respawnDelay = 4;
	ufoDelay = 10;

	// Reset playing field
	asteroids->clearAsteroids();
	ship->lives = 3;
	ship->startShip();
	powerups->reset();
	newLevel();

	currentState = STATE_LEVEL_STARTING;
}

void Game::paintAccuracyRounds() {
	
	if (accuracyRounds <= 0) return;

	// If player hits something with 3/4 of total shots, you get an accuracy round award
	// This paints the award icons in the upper right (yes, it was ripped from q3)

	float dRounds = 60;
	float maxWidth = 250;

	// We want to keep it at 250 max width
	if (dRounds * accuracyRounds > maxWidth)
		dRounds = maxWidth/float(accuracyRounds);
 
        WorldObject::glState->set_active();
	Ship::accuracy->bind();

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glColor3f(1, 1, 1);
	glBegin(GL_QUADS);

	for (int i=0; i < accuracyRounds; i++) {

		float offsetX = (SCREENWIDTH - 5) - dRounds*i;

		glTexCoord2f(0,0); glVertex2f(offsetX, 5+45);
		glTexCoord2f(1,0); glVertex2f(offsetX-45, 5+45);
		glTexCoord2f(1,1); glVertex2f(offsetX-45, 50+45);
		glTexCoord2f(0,1); glVertex2f(offsetX, 50+45);
	}

	glEnd();
	glColor3f(1, 1, 1);
}

void Game::paintLivesRemaining() {

	if (ship->lives <= 0) return;

	float dLives = 60;
	float maxWidth = 250;

	// We want to keep it at 250 max width
	if (dLives * ship->lives > maxWidth)
		dLives = maxWidth/float(ship->lives);

        WorldObject::glState->set_active();

	Ship::shipTexture->bind();

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glColor3f(1, 1, 1);
	glBegin(GL_QUADS);

	for (int i=0; i < ship->lives; i++) {

		float offsetX = (SCREENWIDTH - 5) - dLives*i;

		glTexCoord2f(0,0); glVertex2f(offsetX, 5);
		glTexCoord2f(1,0); glVertex2f(offsetX-45, 5);
		glTexCoord2f(1,1); glVertex2f(offsetX-45, 50);
		glTexCoord2f(0,1); glVertex2f(offsetX, 50);
	}

	glEnd();
	glColor3f(1, 1, 1);
}
