#include "Ship.h"

bool Ship::texturesLoaded = false;
CL_OpenGLSurface 
	*Ship::shipTexture, 
	*Ship::shipThrustingTexture, 
	*Ship::shieldTexture, 
	*Ship::hudCurve, 
	*Ship::hudCurveMask, 
	*Ship::accuracy;

// Constructor
Ship::Ship()
{
	keys[0].set(CL_KEY_SPACE, 150);
	bound.setBoundingCircle(this, 40);
	size = 40.0f;

	lives = 3;

	if (!texturesLoaded) {
		shipTexture = new CL_OpenGLSurface("Ships/ship_1", WorldObject::manager);
		shipThrustingTexture = new CL_OpenGLSurface("Ships/ship_thrust_mask", WorldObject::manager);
		shieldTexture = new CL_OpenGLSurface("Ships/ship_shield", WorldObject::manager);
		hudCurve = new CL_OpenGLSurface("Graphics/hud_curve", WorldObject::manager);
		hudCurveMask = new CL_OpenGLSurface("Graphics/hud_curve_mask", WorldObject::manager);
		accuracy = new CL_OpenGLSurface("Graphics/accuracy", WorldObject::manager);
		texturesLoaded = true;
	}

	exhaust.setRate(40);
	exhaust.setBaseObject(this);
	explosionParticles.setBaseObject(this);
}

// Destructor
Ship::~Ship() 
{
	cout << "Ship Deleted.\n";

	bulletsFired.start();

	while (bulletsFired.next()) {
		delete bulletsFired.get();
	}
}

void Ship::paint( ) {

	// Paint bullets
	bulletsFired.start();

	while (bulletsFired.next())
		bulletsFired.get()->paint();

	// Paint exhaust
	exhaust.paint();

	if (alive) {

		shipTexture->bind();

		// Paint ship
		paintOverlap();

		// Paint engine glow
		if (timeSinceLastAccelerate < .5) {

			shipThrustingTexture->bind();

			float alphaForThrustingMask = (1-timeSinceLastAccelerate)-.5f;

			glColor4f(1, 1, 1, alphaForThrustingMask);
			paintOverlap();
		}

		// Paint shield
		if (timeSinceLastShield < 2.5 || shieldPowerup.active || timeElapsed < 4) {

			float tempSize = size;
			size = 48;

			if (shieldPowerup.active && 

				// Make the shield blink twice before disappearing
				(shieldPowerup.lengthElapsed < shieldPowerup.lengthActive - 1.25
				|| shieldPowerup.lengthElapsed > shieldPowerup.lengthActive - .5)

				&&
				(shieldPowerup.lengthElapsed < shieldPowerup.lengthActive - 2.25
				|| shieldPowerup.lengthElapsed > shieldPowerup.lengthActive - 1.5)) {

				glColor4f(.1f, 1, .1f, .7f);
				timeSinceLastShield = 0;

			} else if (timeElapsed < 1.5) {

				glColor4f(.1f, 1, .1f, .7f);
				timeSinceLastShield = 0;

			} else
				glColor4f(.1f, 1, .1f, (.7f*shieldPower)*(1-(timeSinceLastShield/2.5f)));

			shieldTexture->bind();
			paintOverlap();
			size = tempSize;
		}
	}

	if (!alive && !explosionParticles.hasNoAnimations()) {
		explosionParticles.paint();
	}

	// Paint HUD
	paintHUD();
}

void Ship::update(float timeScale) {

	if (alive) {

		this->timeScale = timeScale;
		exhaust.setInactive();

		// Update position and heading
		x += (velocity*timeScale).getX();
		y += (velocity*timeScale).getY();

		heading += dHeading*timeScale;

		if (dHeading > 360)
			dHeading = 360;

		if (dHeading < -360)
			dHeading = -360;

		// Throttle the variables so the ship comes to a stop eventually
		if (velocity.magnitude != 0 && !accelerating) velocity.magnitude *= 1-(.5f*timeScale);
		if (dHeading != 0 && !turning) dHeading *= 1-(5*timeScale);

		if (*gameState == STATE_LEVEL_PLAYING)
			timeElapsed += timeScale;

		timeSinceLastAccelerate += timeScale;
		timeSinceLastShield += timeScale;
		accelerating = false; turning = false; shield = false;

		handleKey();

		// Ammo/guns
		if (ammoPower <= 1 && !CL_Keyboard::get_keycode(CL_KEY_SPACE))
			ammoPower += .2*timeScale;

		if (gunPowerup.active) {
			gunPowerup.lengthElapsed += timeScale;

			if (gunPowerup.lengthElapsed > gunPowerup.lengthActive) {
				gunPowerup.active = false;
				gunPowerup.lengthElapsed = 0;
			}
		}

		// Shields
		if (shieldPowerup.active) {
			shieldPowerup.lengthElapsed += timeScale;

			if (shieldPowerup.lengthElapsed > shieldPowerup.lengthActive) {
				shieldPowerup.active = false;
				shieldPowerup.lengthElapsed = 0;
				timeSinceLastShield = 0;
			}
		}	

		exhaust.update(timeScale);

		bulletsFired.start();

		while (bulletsFired.next()) {
			bulletsFired.get()->update(timeScale);
		}

		bulletsFired.start();

		while (bulletsFired.next()) {

			if (bulletsFired.get()->isRemovable()) {
				delete bulletsFired.get();
				bulletsFired.remove();
			}
		}

	} else {

		if (!explosionParticles.hasNoAnimations())
			explosionParticles.update(timeScale);

		if (!exhaust.hasNoAnimations())
			exhaust.update(timeScale);
	}

}

void Ship::accelerate() {

	// Add to the velocity vector in the direction the ship is pointing
	velocity.addVector(Vector2(heading, 200*timeScale));
	
	accelerating = true;
	exhaust.setActive();

	timeSinceLastAccelerate = 0;
}

void Ship::turnCW() {
	dHeading += 900*timeScale;
	turning = true;
}

void Ship::turnCCW() {
	dHeading -= 900*timeScale;
	turning = true;
}

void Ship::fireBullet() {

	if (ammoPower >= .1f && *gameState == STATE_LEVEL_PLAYING) {

		if (gunPowerup.active) {

			// If the powerup is active, ship fires three bullets
			Bullet *bullet1, *bullet2, *bullet3;
			
			bullet1 = new Bullet(this, 1, .8f, .8f); bullet2 = new Bullet(this, .8f, .8f, 1); bullet3 = new Bullet(this, 1, .8f, .8f);

			// Modify particle rate to prevent slowdown (lots of trails!)
			bullet1->bulletParticles.setRate(30); bullet2->bulletParticles.setRate(30); bullet3->bulletParticles.setRate(30);

			// Modify direction for a spread pattern
			bullet2->velocity.addVector(Vector2(bullet2->velocity.theta+90, 85));
			bullet3->velocity.addVector(Vector2(bullet3->velocity.theta-90, 85));

			bulletsFired.add(bullet1); bulletsFired.add(bullet2); bulletsFired.add(bullet3);

			shots+=3;

		} else {
			bulletsFired.add(new Bullet(this, .8f, .8f, 1));
			shots++;
		}

		ammoPower -= .1f;
	}
}

SimpleList<Bullet> *Ship::getBulletsList() {
	return &bulletsFired;
}

void Ship::isHit() {

	if (shieldPowerup.active) return;
	if (timeElapsed < 4) return;

	shieldPower -= .2f;

	// Game over
	if (shieldPower < 0) {
		alive = false;
		explosionParticles.burst(50);
		exhaust.setInactive();
		clearBulletList();
	}

	timeSinceLastShield = 0;

	cout << "Shieldpower: " << shieldPower << "\n";
}

void Ship::bounceOffObject(WorldObject *objectHit, int deflection) {
	
	Vector2 rayFromObjectCenter;
	rayFromObjectCenter.setXY(objectHit->x-x, objectHit->y-y);
	velocity.addVector(Vector2(rayFromObjectCenter.theta+180, deflection));
}


void Ship::handleKey()
{
	//cout << "Handling Key\n";

	if (CL_Keyboard::get_keycode(CL_KEY_RIGHT)) {
		turnCW();
	}

	if (CL_Keyboard::get_keycode(CL_KEY_LEFT)) {
		turnCCW();
	}

	if (CL_Keyboard::get_keycode(CL_KEY_UP)) {
		accelerate();
	}

	//if (CL_Keyboard::get_keycode(CL_KEY_DOWN)) {
		//shieldUp();
	//}

	if (keys[0].isPressed()) {
		fireBullet();
	}
}

void Ship::clearBulletList() {

	bulletsFired.start();

	while (bulletsFired.next()) {

		delete bulletsFired.get();
		bulletsFired.remove();
	}
}

void Ship::paintHUD() {

	int hudSize = 256;

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	float shieldHeight = (1-shieldPower)*float(hudSize);
	float powerHeight = (1-ammoPower)*float(hudSize);

	// Draw the color bases
	hudCurveMask->bind();

	glBegin(GL_QUADS);	

		glColor3f(0, 1, 0);
	
		glTexCoord2f(0,1-shieldPower); glVertex2f(-hudSize/4, -hudSize+SCREENHEIGHT+shieldHeight);
		glTexCoord2f(1,1-shieldPower); glVertex2f(hudSize-hudSize/4, -hudSize+SCREENHEIGHT+shieldHeight);
		glTexCoord2f(1,1); glVertex2f(hudSize-hudSize/4, SCREENHEIGHT);
		glTexCoord2f(0,1); glVertex2f(-hudSize/4   , SCREENHEIGHT);

		glColor3f(1, 0, 0);
									
		glTexCoord2f(1,1-ammoPower); glVertex2f(SCREENWIDTH-hudSize+hudSize/4, -hudSize+SCREENHEIGHT+powerHeight);
		glTexCoord2f(0,1-ammoPower);  glVertex2f(SCREENWIDTH+hudSize/4, -hudSize+SCREENHEIGHT+powerHeight);
		glTexCoord2f(0,1); glVertex2f(SCREENWIDTH+hudSize/4, SCREENHEIGHT);
		glTexCoord2f(1,1); glVertex2f(SCREENWIDTH-hudSize+hudSize/4, SCREENHEIGHT);

	glEnd();

	// Draw overlay curves
	hudCurve->bind();

	glBegin(GL_QUADS);	

		glColor3f(.6f, 1, .6f);
	
		glTexCoord2f(0,0); glVertex2f(-hudSize/4, -hudSize+SCREENHEIGHT);
		glTexCoord2f(1,0); glVertex2f(hudSize-hudSize/4, -hudSize+SCREENHEIGHT);
		glTexCoord2f(1,1); glVertex2f(hudSize-hudSize/4, SCREENHEIGHT);
		glTexCoord2f(0,1); glVertex2f(-hudSize/4   , SCREENHEIGHT);

		glColor3f(1, 1, 1);
									
		glTexCoord2f(1,0); glVertex2f(SCREENWIDTH-hudSize+hudSize/4, -hudSize+SCREENHEIGHT);
		glTexCoord2f(0,0); glVertex2f(SCREENWIDTH+hudSize/4, -hudSize+SCREENHEIGHT);
		glTexCoord2f(0,1); glVertex2f(SCREENWIDTH+hudSize/4, SCREENHEIGHT);
		glTexCoord2f(1,1); glVertex2f(SCREENWIDTH-hudSize+hudSize/4, SCREENHEIGHT);

	glEnd();

	glColor3f(1, 1, 1);

        int height = Game::font->get_height();
	Game::font->draw(6, SCREENHEIGHT-height, "Shields");
        std:string ammoString = "Ammo";
        int width = Game::font->get_width(ammoString);
	Game::font->draw(SCREENWIDTH-width, SCREENHEIGHT-height, ammoString);
}

void Ship::startShip() {

	lives--;
	hits = 0, shots = 0;
	timeSinceLastAccelerate = 10;
	timeSinceLastShield = 10;
	timeElapsed = 0;
	timeElapsing = false;
	accelerating = false; turning = false; shield = false; alive = true;
	heading = 0.0f, dHeading = 0.0f; shieldPower = 1.0f; ammoPower = 1.0f;
	velocity = Vector2(0, 0);
	x = -200; y = 0;

	//findSafeStart();

	// Powerups
	shieldPowerup.lengthActive = 10;	gunPowerup.lengthActive = 10;	lifePowerup.lengthActive = 0;
	shieldPowerup.active = false;		gunPowerup.active = false;		lifePowerup.active = false;
	shieldPowerup.lengthElapsed = 0;	gunPowerup.lengthElapsed = 0;	lifePowerup.lengthElapsed = 0;
}

void Ship::bulletHitSomething() {
	hits++;
}
